(***************************************************************************
  HC unit
  A parser for compiling help files
  Copyright Peter Brandstrom

  Written by Peter Brandstrom, January 1994

***************************************************************************)
unit HC;
{$B-,X+}

interface

  uses
    Dos, Objects, Strings,
    OPString,
    HelpFiles, THelp,
    TextStrm;

  const
    AliasXRefDirective            = 'ALIASXREFCHAR';
    AutoWrapDirective             = 'AUTOWRAP';
    BeginXRefDirective            = 'BEGINXREFCHAR';
    CommentDirective              = 'COMMENTCHAR';
    EndXRefDirective              = 'ENDXREFCHAR';
    FixedDirective                = 'FIXED';
    HighlightDirective            = 'HIGHLIGHTCHAR';
    IncludeDirective              = 'INCLUDE';
    IndexDirective                = 'INDEX';
    IndexTopicNamesDirective      = 'INDEXTOPICNAMES';
    MainTopicDirective            = 'MAINTOPIC';
    NoWrapDirective               = 'NOWRAP';
    QuoteDirective                = 'QUOTECHAR';
    RestoreStateDirective         = 'RESTORESTATE';
    SaveStateDirective            = 'SAVESTATE';
    SignatureDirective            = 'SIGNATURE';
    StampDirective                = 'STAMP';
    SubIndexDirective             = 'SUBINDEX';
    TopicDirective                = 'TOPIC';
    WrapDirective                 = 'WRAP';


    errNoTopicDefined             = 'No topic defined yet';
    errMainTopicRedefined         = 'Main topic already defined';
    errUnresolvedRef              = 'Unresolved reference to ';
    errCharacterExpected          = 'Character expected';
    errOnOffExpected              = 'ON or OFF expected';
    errNoPreviousState            = 'No previous state to restore to';
    errUnrecognizedDirective      = 'Unrecognized compiler directive ';
    errLineTooLong                = 'Line too long';
    errFile                       = 'File ';
    errNotFound                   = ' not found';
    errAlreadyIncluded            = ' already included in compilation';
    errUnterminatedCrossRef       = 'Unterminated topic reference';
    errTopicNameExpected          = 'Topic name expected';
    errRedefinitionOf             = 'Redefinition of ';


  type
    PHelpCState = ^THelpCState;
    THelpCState =
      object
        { This object holds all user configurable values which
           are saved and restored by .SAVESTATE and .RESTORESTATE }
        BeginXRefChar    : Char;
        AliasXRefChar    : Char;
        EndXRefChar      : Char;
        QuoteChar        : Char;
        CommentChar      : Char;
        HighlightChar    : Char;

        Wrap             : Boolean;
        AutoWrap         : Boolean;
        IndexTopicNames  : Boolean;

        FallBack         : String;

        Next             : PHelpCState;

        constructor Init(const Data:THelpCState; aNext:PHelpCState);
        function PopState(var State:THelpCState):PHelpCState;
      end;


  const
    StartState : THelpCState =
      (BeginXRefChar   : '{';
       AliasXRefChar   : ':';
       EndXRefChar     : '}';
       QuoteChar       : '\';
       CommentChar     : ';';
       HighlightChar   : '#';
       Wrap            : True;
       AutoWrap        : True;
       IndexTopicNames : True);


    StartingTopicNumber  = 2;

  type
    TopicNumberType  = Word;
    LineNumberType   = Word;

    PFilePos = ^TFilePos;
    TFilePos =
      object(TObject)
        { A file position }
        Name    : PString;
        FileID  : Integer;
        Line    : LineNumberType;
        constructor Init(const aName:String; aFileID:Integer; aLine:LineNumberType);
        destructor Done; virtual;
      end;

    PTopicName = ^TTopicName;
    TTopicName =
      object(TFilePos)
        { Position of a topic definition in a file }
        Context : TopicNumberType;
        constructor Init(const aName:String; aContext:TopicNumberType;
          aFileID:Integer; aLine:LineNumberType);
      end;

    PTopicNameList = ^TTopicNameList;
    TTopicNameList =
      object (TSortedCollection)
        { List of topic definitions }
        CaseSensitive : Boolean;
          { True if topics are case sensitive }
        function Compare(Key1, Key2:Pointer):Integer; virtual;
        function GetTopicNumber(const Topic:String; var Number:TopicNumberType):Boolean;
        function KeyOf(Item:Pointer):Pointer; virtual;
      end;

    PHelpCompiler = ^THelpCompiler;
    THelpCompiler =
      object (TObject)
        { Compiles help text into TPH files }
        HelpText         : PTextStream;
        HelpName         : String;
        FileID           : Integer;
        LineNumber       : LineNumberType;

        HelpFile         : PHelpFile;
        CurTopic         : PTopic;
        CurTopicName     : String;
        TopicIsIndexed   : Boolean;
        TopicList        : PTopicNameList;
        GotMainTopic     : Boolean;
        MainTopic        : TopicNumberType;

        State            : THelpCState;
        SavedStateList   : PHelpCState;
        FilesCompiled    : TCollection;
        IncludePath      : String;

        Pass2            : Boolean;
        Line             : String;
        BlankLineCount   : Integer;

        TopicNumber      : TopicNumberType;
        TopicsToCompile  : Integer;
        TopicsCompiled   : Integer;

        BlankLinesAtEndOfTopic : Integer;

        constructor Init(AFileName:String; aHelpText:PTextStream;
                    aHelpFile:PHelpFile; aIncludePath:String);
        destructor Done; virtual;
        procedure AddCrossRef(const LinkText, Alias:String);
        procedure Compile;
        procedure Directive;
        procedure Error(const Msg:String);
        procedure FinishTopic;
        procedure GetLine;
        function  GetTopicNumber:TopicNumberType;
        procedure Include(const aFileName:String);
        procedure Message(const MsgType, FileName:String; aLine:Word; const Msg:String); virtual;
        procedure Parse1;
        procedure Parse2;
        procedure Progress; virtual;
        procedure ScanCrossRefs;
        procedure TopicDefinition;
      end;


(***************************************************************************
***************************************************************************)
implementation


  function GetWord(var P:PChar):String;
    { Return the word at P^.  A word ends with a new line or whitespace }
    var
      i : Integer;
  begin
    i:=0;

    while not (P^ in [#0,' ',^I]) do
    begin
      Inc(i);
      GetWord[i]:=P^;
      Inc(P);
    end;

    GetWord[0]:=Chr(i);
  end;


  function Match(var P:PChar; const S:String):Boolean;
    { Return true if S can be found at P^. Case insensitive, whole words }
  begin
    Match:=False;
    if (StrLen(P)>=Length(S)) and (StrLIComp(P, @S[1], Length(S))=0) and
      ((P+Length(S))^ in [#0,' ',^I]) then
    begin
      Match:=True;
      Inc(P, Length(S));
    end;
  end;


  procedure Scan(var Source:PChar; SEnd:PChar; var Dest:PChar; Ch1, Ch2, Quote:Char); assembler;
    { Copy Source to Dest, stopping if #0, Ch1 or Ch2 encountered.
      Quote is a quoting character }
  asm
      push ds
      cld

      lds  si,Dest
      les  di,[si]

      lds  si,Source
      lds  si,[si]

      mov  ah,Quote
      mov  bl,Ch1
      mov  bh,Ch2
      jmp  @Start

    @Loop:
      stosb

    @Start:
      lodsb

      or   al,al
      je   @Fin

      cmp  al,ah
      je   @Quote

      cmp  al,bl
      je   @Fin
      cmp  al,bh
      jne  @Loop

      jmp  @Fin

    @Quote:
      cmp  si,SEnd.Word
      jae  @Fin
      movsb
      jmp  @Start


    @Fin:
      mov  ax,si
      dec  ax
      lds  si,Source
      mov  [si],ax

      lds  si,Dest
      mov  [si],di

      pop  ds
  end;


  procedure ScanToBuf(const Line:String; var LineP:PChar; var Buf:String; Ch1, Ch2, QCh:Char);
    { A #Scan# procedure wrapper, more String oriented }
    var
      BufP : PChar;
  begin
    BufP:=@Buf[1];
    Scan(LineP, @Line[Length(Line)+1], BufP, Ch1, Ch2, QCh);
    Byte(Buf[0]):=BufP-@Buf[1];
  end;


  procedure SkipBlanks(var S:PChar);
    { Skip tabs and spaces }
  begin
    while S^ in [' ',^I] do
      Inc(S);
  end;


  function Trim(const S:String):String;
    { Remove leading and trailing tabs and spaces }
    var
      i, j : Integer;
  begin
    i:=1;
    while (i<Length(S)) and (S[i] in [' ',^I]) do
      Inc(i);

    j:=Length(S);
    while (j>0) and (S[j] in [' ',^I]) do
      Dec(j);

    Trim:=Copy(S, i, j+1-i);
  end;


(***************************************************************************
***************************************************************************)

  constructor TFilePos.Init(const aName:String; aFileID:Integer; aLine:LineNumberType);
    { Create a file position record }
  begin
    Name:=NewStr(aName);
    FileID:=aFileID;
    Line:=aLine;
  end;

  destructor TFilePos.Done;
    { Dispose of the name }
  begin
    DisposeStr(Name);
  end;


(***************************************************************************
***************************************************************************)

  constructor TTopicName.Init(const aName:String; aContext:TopicNumberType;
                        aFileID:Integer; aLine:LineNumberType);
    { Create a topic entry }
  begin
    inherited Init(aName, aFileID, aLine);
    Context:=aContext;
  end;


(***************************************************************************
***************************************************************************)

  function TTopicNameList.Compare(Key1, Key2:Pointer):Integer;
    { Case insensitive topic name order }
  begin
    if CaseSensitive then
      Compare:=Ord(CompString(PString(Key1)^, PString(Key2)^))-1
    else
      Compare:=Ord(CompUCString(PString(Key1)^, PString(Key2)^))-1;
  end;


  function TTopicNameList.GetTopicNumber(const Topic:String; var Number:TopicNumberType):Boolean;
    { Return topic number of a topic }
    var
      Index : Integer;
  begin
    GetTopicNumber:=True;
    if Search(@Topic, Index) then
      Number:=PTopicName(At(Index))^.Context
    else
      GetTopicNumber:=False;
  end;


  function TTopicNameList.KeyOf(Item:Pointer):Pointer;
    { Use the name as the search key }
  begin
    KeyOf:=PTopicName(Item)^.Name;
  end;


(***************************************************************************
***************************************************************************)

  constructor THelpCState.Init(const Data:THelpCState; aNext:PHelpCState);
    { Shorthand constructor to link help states easily. Saves the
      current state in a linked list }
  begin
    Self:=Data;
    Next:=aNext;
  end;

  function THelpCState.PopState(var State:THelpCState):PHelpCState;
    { Shorthand destructor to unlink help states easily. Restores the
      previous compiler state from the linked list and disposes of it }
  begin
    PopState:=Next;
    State:=Self;
    Dispose(@Self);
  end;


(***************************************************************************
***************************************************************************)

  constructor THelpCompiler.Init;
    { Initialize help compiler }
  begin
    inherited Init;

    HelpName:=AFileName;
    State.FallBack:=AFileName;
    HelpText:=aHelpText;
    HelpFile:=aHelpFile;
    IncludePath:=aIncludePath;
    BlankLinesAtEndOfTopic:=1;

    New(TopicList, Init(200,200));
    FilesCompiled.Init(5,5);
  end;


  destructor THelpCompiler.Done;
    { Clean up }
  begin
    Dispose(HelpText, Done);
    Dispose(TopicList, Done);
    FilesCompiled.Done;
    Dispose(HelpFile, Done);

    while SavedStateList<>Nil do
      SavedStateList:=SavedStateList^.PopState(State);
  end;


  procedure THelpCompiler.AddCrossRef;
    { Add a cross reference.  Checks if it is an unresolved reference }
    var
      Number : TopicNumberType;
  begin
    if TopicList^.GetTopicNumber(Alias, Number) then
      if CurTopic<>Nil then
        CurTopic^.WriteKeyword(LinkText, Number)
      else
        Error(errNoTopicDefined)
    else
    begin
      CurTopic^.Write(LinkText);
      Error(errUnresolvedRef + Alias);
    end;
  end;


  procedure THelpCompiler.Compile;
    { Compile the help file in two passes }
  begin
    FilesCompiled.Insert(New(PFilePos, Init(HelpName, 0, 0)));
    State.Init(StartState, Nil);

    TopicNumber:=StartingTopicNumber;
    Pass2:=False;
    Progress;
    Parse1;
    Pass2:=True;

    HelpText^.Reset;
    HelpText^.Seek(0);
    LineNumber:=0;
    State.Init(StartState, Nil);
    FilesCompiled.FreeAll;
    FilesCompiled.Insert(New(PFilePos, Init(HelpName, 0, 0)));

    Progress;
    TopicNumber:=StartingTopicNumber;
    Parse2;
  end;


  procedure THelpCompiler.Directive;
    { Parse all compiler directives }
    var
      LineP   : PChar;
      StrArg  : String;

    function StringArgDirective(const Command:String):Boolean;
    begin
      StringArgDirective:=False;
      if Match(LineP, Command) then
      begin
        SkipBlanks(LineP);
        StrArg:=StrPas(LineP);
        StringArgDirective:=True;
      end;
    end;

    function DefChar(const Command:String; var Ch:Char):Boolean;
    begin
      DefChar:=False;

      if StringArgDirective(Command) then
      begin
        if StrArg='' then
          Error(errCharacterExpected)
        else
          Ch:=StrArg[1];

        DefChar:=True;
      end
    end;

    function Switch(const Command:String; var Flag:Boolean):Boolean;
    begin
      Switch:=False;

      if Match(LineP, Command) then
      begin
        SkipBlanks(LineP);
        if Match(LineP, 'ON') then
          Flag:=True
        else
          if Match(LineP, 'OFF') then
            Flag:=False
          else
            Error(errOnOffExpected);

        Switch:=True;
      end;
    end;

  begin
    LineP:=@Line[2];

    if Match(LineP, TopicDirective) then
      TopicDefinition
    else if StringArgDirective(IncludeDirective) then
      Include(StrArg)
    else
    if not DefChar(QuoteDirective, State.QuoteChar) then
      if Pass2 then
        { Indexing --------------------------------------}
        if StringArgDirective(IndexDirective) then
        begin
          if CurTopic<>Nil then
          begin
            HelpFile^.Index^.AddItem(StrArg, State.FallBack, CurTopic^.TopicNum);
            TopicIsIndexed:=True
          end
          else
            Error(errNoTopicDefined);
        end
        else if StringArgDirective(SubIndexDirective) then
          State.FallBack:=StrArg
        { TPH specifics --------------------------------------}
        else if StringArgDirective(SignatureDirective) then
          PBorHelpFile(HelpFile)^.SetSignature(StrArg)
        else if StringArgDirective(StampDirective) then
          PBorHelpFile(HelpFile)^.SetStamp(StrArg)
        else if Match(LineP, MainTopicDirective) then
          if CurTopic=Nil then
            Error(errNoTopicDefined)
          else
            if GotMainTopic then
              Error(errMainTopicRedefined)
            else
            begin
              GotMainTopic:=True;
              HelpFile^.SetMainTopic(CurTopic^.TopicNum)
            end
        { wrapping --------------------------------------}
        else if Match(LineP, WrapDirective) then
        begin
          State.AutoWrap:=False;
          State.Wrap:=True;
          if CurTopic^.FixedLines then
            CurTopic^.ToggleFixed;
        end
        else if Match(LineP, NoWrapDirective) or
                Match(LineP, FixedDirective) then
        begin
          State.AutoWrap:=False;
          State.Wrap:=False;
          if not CurTopic^.FixedLines then
            CurTopic^.ToggleFixed;
        end
        else if Match(LineP, AutoWrapDirective) then
          State.AutoWrap:=True
        { state --------------------------------------}
        else if Match(LineP, SaveStateDirective) then
        begin
          if CurTopic<>Nil then
            State.Wrap:=not CurTopic^.FixedLines;
          New(SavedStateList, Init(State, SavedStateList))
        end
        else if Match(LineP, RestoreStateDirective) then
          if SavedStateList=Nil then
            Error(errNoPreviousState)
          else
          begin
            SavedStateList:=SavedStateList^.PopState(State);
            if (CurTopic<>Nil) and (CurTopic^.FixedLines=State.Wrap) then
              CurTopic^.ToggleFixed;
          end
        else
        { characters --------------------------------------}
        if not DefChar(BeginXRefDirective, State.BeginXRefChar) and
           not DefChar(EndXRefDirective, State.EndXRefChar) and
           not DefChar(QuoteDirective, State.QuoteChar) and
           not DefChar(AliasXRefDirective, State.AliasXRefChar) and
           not DefChar(CommentDirective, State.CommentChar) and
           not DefChar(HighlightDirective, State.HighlightChar) and

        { switches --------------------------------------}
           not Switch(IndexTopicNamesDirective, State.IndexTopicNames) then

          Error(errUnrecognizedDirective + stUpcase(GetWord(LineP)));
  end;


  procedure THelpCompiler.Error;
    { Output an error message }
  begin
    Message('Error: ', HelpName, LineNumber, Msg);
  end;


  procedure THelpCompiler.FinishTopic;
    { Flush topic }
    var
      i : Integer;
  begin
    if CurTopic<>Nil then
    begin
      if State.IndexTopicNames and not TopicIsIndexed then
        HelpFile^.Index^.AddItem(CurTopicName, State.FallBack, CurTopic^.TopicNum);

      for i:=1 to BlankLinesAtEndOfTopic do
        CurTopic^.BlankLine;

      if TopicsCompiled=0 then
        HelpFile^.SetMainTopic(CurTopic^.TopicNum);

      HelpFile^.AddTopic(CurTopic);
      CurTopic:=Nil;

      Inc(TopicsCompiled);
      BlankLineCount:=0;
      Progress;
    end;
  end;


  procedure THelpCompiler.GetLine;
    { Retrieve next non comment line }
  begin
    repeat
      HelpText^.ReadLine(Line);
      Inc(LineNumber);
    until (Length(Line)=0) or (Line[1]<>State.CommentChar);

    if Length(Line)=255 then
    begin
      Error(errLineTooLong);
      Dec(Line[0]);
    end;

    while (Line<>'') and (Line[Length(Line)] in [' ',^I]) do
      Dec(Line[0]);

    Line[Length(Line)+1]:=#0
  end;


  function THelpCompiler.GetTopicNumber;
    { Return the succeeding topic number }
  begin
    GetTopicNumber:=TopicNumber;
    Inc(TopicNumber);
  end;


  procedure THelpCompiler.Include(const aFileName:String);
    { Include a file into the compilation }

    var
      FileName : String;

    function CheckIfAlreadyIncluded(P:PFilePos):Boolean; far;
    begin
      CheckIfAlreadyIncluded:=CompUCString(FExpand(P^.Name^), FileName)=Equal;
    end;

    var
      OldSource   : PTextStream;
      OldLine     : LineNumberType;
      OldHelpName : String;
      OldFileID   : Integer;

  begin
    FileName:=FExpand(aFileName);

    if FilesCompiled.FirstThat(@CheckIfAlreadyIncluded)=Nil then
    begin
      OldLine:=LineNumber;
      OldHelpName:=HelpName;
      OldSource:=HelpText;
      OldFileID:=FileID;

      New(HelpText, Init(FSearch(aFileName, IncludePath), stOpenRead, 2048));
      if HelpText^.Status=stOK then
      begin
        HelpName:=aFileName;

        FilesCompiled.Insert(New(PFilePos, Init(HelpName, FileID, LineNumber)));
        FileID:=FilesCompiled.Count-1;
        LineNumber:=0;

        if Pass2 then
          Parse2
        else
          Parse1;

        FileID:=OldFileID;
      end
      else
        if not Pass2 then
          Error(errFile + aFileName + errNotFound);

      Dispose(HelpText, Done);

      LineNumber:=OldLine;
      HelpName:=OldHelpName;
      HelpText:=OldSource;
    end
    else
      if not Pass2 then
        Error(errFile + stUpcase(aFileName) + errAlreadyIncluded);
  end;


  procedure THelpCompiler.Message;
    { Output a message, used for errors }
  begin
  end;


  procedure THelpCompiler.Parse1;
    { Parse the help file, first pass.  Only look for topic definitions }
  begin
    if HelpText^.Status<>stOK then
      Error(errFile + HelpName + errNotFound)
    else
    begin
      GetLine;

      while HelpText^.Status = stOK do
      begin
        if Line[1]='.' then
          Directive;

        GetLine;
      end;
    end;
  end;


  procedure THelpCompiler.Parse2;
    { Parse the help file, second pass }
  begin
    if HelpText^.Status=stOK then
    begin
      GetLine;

      while HelpText^.Status = stOK do
      begin
        if Line[1]='.' then
          Directive
        else
          if CurTopic=Nil then
          begin
            if Line<>'' then
              Error(errNoTopicDefined);
          end
          else
            if Line='' then
              Inc(BlankLineCount)
            else
            begin
              while BlankLineCount<>0 do
              begin
                CurTopic^.BlankLine;
                Dec(BlankLineCount);
              end;
              ScanCrossRefs;
            end;

        GetLine;
      end;

      FinishTopic;
    end;
  end;


  procedure THelpCompiler.Progress;
    { Progress indication. Called when Pass 1 and Pass 2 begins, and
      after every topic compiled }
  begin
  end;


  procedure THelpCompiler.ScanCrossRefs;
    { Search for cross references and add them to the current topic }
    var
      LineP   : PChar;
      Buf     : String;
      Keyword : String;

    procedure PrintBuf;
    begin
      if Buf<>'' then
      begin
        if not State.AutoWrap and CurTopic^.StartOfLine and (Buf[1]=' ') then
          Buf[1]:=#255;
        CurTopic^.Write(Buf);
        Buf:='';
      end;
    end;

  begin
    Buf:='';
    LineP:=@Line[1];

    while LineP^<>#0 do
    begin
      ScanToBuf(Line, LineP, Buf, State.BeginXRefChar,
                State.HighlightChar, State.QuoteChar);

      if LineP^=State.HighlightChar then
      begin
        PrintBuf;
        Inc(LineP);
        CurTopic^.ToggleMarked;
      end
      else
        if LineP^=State.BeginXRefChar then
        begin
          PrintBuf;

          Inc(LineP);
          ScanToBuf(Line, LineP, Buf, State.AliasXRefChar,
                    State.EndXRefChar, State.QuoteChar);

          Keyword:=Buf;

          if LineP^=State.AliasXRefChar then
          begin
            Inc(LineP);
            ScanToBuf(Line, LineP, Buf, State.EndXRefChar, #0, State.QuoteChar);
          end;

          if LineP^=#0 then
            Error(errUnterminatedCrossRef)
          else
          begin
            AddCrossRef(Keyword, Trim(Buf));

            Buf:='';
            Inc(LineP);
          end;
        end;
    end;

    if not State.AutoWrap and CurTopic^.StartOfLine and
         (Buf<>'') and (Buf[1]=' ') then
      Buf[1]:=#255;

    CurTopic^.WriteLn(Buf);
  end;



  procedure THelpCompiler.TopicDefinition;
    { Parse a topic definition }
    var
      Start         : PChar;
      Number        : TopicNumberType;
      PreviousTopic : PTopicName;
      Index         : Integer;
      FirstTime     : Boolean;
      LineP         : PChar;
  begin
    if Pass2 then
      FinishTopic;

    FirstTime:=True;
    TopicIsIndexed:=False;

    LineP:=@Line[7];
    SkipBlanks(LineP);

    repeat
      ScanToBuf(Line, LineP, CurTopicName, ',', '=', State.QuoteChar);

      if CurTopicName='' then
      begin
        Error(errTopicNameExpected);
        Exit;
      end;

      CurTopicName:=Trim(CurTopicName);

      if LineP^ = '=' then
      begin
        Inc(LineP);
        TopicNumber:=0;
        while LineP^ in ['0'..'9'] do
        begin
          TopicNumber:=10*TopicNumber + Ord(LineP^) - Ord('0');
          Inc(LineP);
        end;
      end;

      if LineP^ = ',' then
        Inc(LineP);

      if FirstTime then
      begin
        FirstTime:=False;
        Number:=GetTopicNumber;
      end
      else
        GetTopicNumber;

      if not Pass2 then
        if TopicList^.Search(@CurTopicName, Index) then
        begin
          PreviousTopic:=TopicList^.At(Index);
          Message('Error: ',
                  PFilePos(FilesCompiled.At(PreviousTopic^.FileID))^.Name^,
                  PreviousTopic^.Line,
                  'First definition of ' + CurTopicName);

          Error(errRedefinitionOf + CurTopicName);
        end
        else
          TopicList^.Insert(New(PTopicName, Init(CurTopicName, Number, FileID, LineNumber)));

      SkipBlanks(LineP);
    until LineP^=#0;

    if Pass2 then
    begin
      CurTopic:=HelpFile^.NewTopic(Number, Nil);
      if CurTopic^.FixedLines=State.Wrap then
        CurTopic^.ToggleFixed;
    end
    else
      Inc(TopicsToCompile);
  end;


end.